#include <string.h>
#include "lhsrhs.h"
#include "global.h"
#include "mechprint.h"
#include "gtopology.h"
#include "dloadcase.h"
#include "inicd.h"


/**
  Constructor initializes data members to zero or default values.

  Created by JK,
*/
lhsrhs::lhsrhs ()
{
  //  number of DOFs
  ndof=Ndofm;
  //  number of load case
  nlc=Mb->nlc;
  
  //  array containing unknowns/DOFs
  lhs=NULL;
  //  array containing first time derivatives of unknowns/DOFs
  tdlhs=NULL;
  //  array containing second derivatives of unknowns/DOFs
  stdlhs=NULL;
  //  array containing the right hand side
  rhs=NULL;
  //  array containing eigenvalues
  w=NULL;
}




/**
  Destructor releases allocated memory of the lhsrhs object.

  Created by JK,
*/
lhsrhs::~lhsrhs ()
{
  delete [] lhs;
  delete [] tdlhs;
  delete [] stdlhs;
  delete [] rhs;
  delete [] w;
}



/**
  Function allocates arrays for left and right hand sides of problem.
   
  @return The function does not return anything.

  Created by JK, 29.6.2001
*/
void lhsrhs::alloc ()
{
  switch (Mp->tprob){
  case linear_statics:{
    lhs = new double [nlc*ndof];
    memset (lhs,0,(nlc*ndof)*sizeof(double));
    rhs = new double [nlc*ndof];
    memset (rhs,0,(nlc*ndof)*sizeof(double));
    break;
  }
  case eigen_dynamics:{
    lhs = new double [Mp->eigsol.nev*ndof];
    memset (lhs,0,(Mp->eigsol.nev*ndof)*sizeof(double));
    
    switch (Mp->eigsol.teigsol){
    case inv_iteration:{
      w = new double [1];
      break;
    }
    case subspace_it_jacobi:{
      w = new double [Mp->eigsol.nev];
      break;
    }
    case subspace_it_gsortho:{
      w = new double [Mp->eigsol.nev];
      break;
    }
    default:{
      print_err("unknown eigenvalue solver is required", __FILE__, __LINE__, __func__);
    }
    }

    break;
  }
  case forced_dynamics:{
    if (Mp->tforvib==modal_analysis){
      lhs = new double [ndof+Mp->eigsol.nev*ndof];
      memset (lhs,0,(ndof+Mp->eigsol.nev*ndof)*sizeof(double));

      switch (Mp->eigsol.teigsol){
      case inv_iteration:{
	w = new double [1];
	break;
      }
      case subspace_it_jacobi:{
	w = new double [Mp->eigsol.nev];
	break;
      }
      case subspace_it_gsortho:{
	w = new double [Mp->eigsol.nev];
	break;
      }
      default:{
	print_err("unknown eigenvalue solver is required", __FILE__, __LINE__, __func__);
      }
      }
      
    }
    else{
      lhs = new double [nlc*ndof];
      memset (lhs,0,(nlc*ndof)*sizeof(double));
      tdlhs = new double [nlc*ndof];
      memset (tdlhs,0,(nlc*ndof)*sizeof(double));
      stdlhs = new double [nlc*ndof];
      memset (stdlhs,0,(nlc*ndof)*sizeof(double));
      rhs = new double [nlc*2*ndof];
      memset (rhs,0,(nlc*2*ndof)*sizeof(double));
    }
    break;
  }
  case mat_nonlinear_statics:{
    lhs = new double [nlc*ndof];
    memset (lhs,0,(nlc*ndof)*sizeof(double));
    rhs = new double [nlc*2*ndof];
    memset (rhs,0,(nlc*2*ndof)*sizeof(double));
    break;
  }
  case geom_nonlinear_statics:{
    lhs = new double [nlc*ndof];
    memset (lhs,0,(nlc*ndof)*sizeof(double));
    rhs = new double [nlc*2*ndof];
    memset (rhs,0,(nlc*2*ndof)*sizeof(double));
    break;
  }
  case mech_timedependent_prob:{
    lhs = new double [nlc*ndof];
    memset (lhs,0,(nlc*ndof)*sizeof(double));
    rhs = new double [nlc*ndof];
    memset (rhs,0,(nlc*ndof)*sizeof(double));
    break;
  }
  case growing_mech_structure:{
    lhs = new double [ndof];
    memset (lhs,0,(ndof)*sizeof(double));
    rhs = new double [ndof];
    memset (rhs,0,(ndof)*sizeof(double));
    break;
  }
  case earth_pressure:{
    lhs = new double [nlc*ndof];
    memset (lhs,0,(nlc*ndof)*sizeof(double));
    rhs = new double [nlc*2*ndof];
    memset (rhs,0,(nlc*2*ndof)*sizeof(double));
    break;
  }
  case layered_linear_statics:{    lhs = new double [nlc*ndof];
    memset (lhs,0,(nlc*ndof)*sizeof(double));
    rhs = new double [nlc*ndof];
    memset (rhs,0,(nlc*ndof)*sizeof(double));
    break;
  }
  case lin_floating_subdomain:{
    lhs = new double [nlc*ndof];
    memset (lhs,0,(nlc*ndof)*sizeof(double));
    rhs = new double [nlc*ndof];
    memset (rhs,0,(nlc*ndof)*sizeof(double));
    break;
  }
  case nonlin_floating_subdomain:{
    lhs = new double [nlc*ndof];
    memset (lhs,0,(nlc*ndof)*sizeof(double));
    rhs = new double [2*nlc*ndof];
    memset (rhs,0,(2*nlc*ndof)*sizeof(double));
    break;
  }
  case hemivar_inequal:{
    lhs = new double [nlc*ndof];
    memset (lhs,0,(nlc*ndof)*sizeof(double));
    rhs = new double [nlc*ndof];
    memset (rhs,0,(nlc*ndof)*sizeof(double));
    break;
  }
  case load_balancing:{
    lhs = new double [nlc*ndof];
    memset (lhs,0,(nlc*ndof)*sizeof(double));
    rhs = new double [nlc*ndof];
    memset (rhs,0,(nlc*ndof)*sizeof(double));
    break;
  }
    
  default:{
    print_err("unknown problem type is required", __FILE__, __LINE__, __func__);
  }
  }

}



/**
  Function returns pointer to %vector of unknowns (lhs %vector).
  This %vector contains nodal values.
   
  @param i - load case id
   
  @return The function returns pointer to the array containing %vetcor of unknowns.

  Created by JK,
*/
double *lhsrhs::give_lhs (long i)
{
  return (lhs+i*ndof);
}



/**
  Function returns pointer to %vector of first time derivatives of unknowns (tdlhs %vector).
  This %vector contains first time derivatives of nodal values.
   
  @param i - load case id
   
  @return The function returns pointer to the array containing %vetcor of first time drivatives of unknowns.

  Created by JK,
*/
double *lhsrhs::give_tdlhs (long i)
{
  return (tdlhs+i*ndof);
}



/**
  Function returns pointer to %vector of second time derivatives of unknowns (stdlhs %vector)
  this %vector contains second time derivatives of nodal values
   
  @param i - load case id
   
  @return The function returns pointer to the array containing %vetcor of second time drivatives of unknowns.

  Created by JK,
*/
double *lhsrhs::give_stdlhs (long i)
{
  return (stdlhs+i*ndof);
}



/**
  Function returns pointer to %vector of right hand side (rhs %vector)
  This %vector contains prescribed nodal values.
   
  @param i - load case id
   
  @return The function returns pointer to the array containing %vetcor of precribed nodal values.

  Created by JK,
*/
double *lhsrhs::give_rhs (long i)
{
  return (rhs+i*ndof);
}



/**
  Old version of text results output. The function prints displacements at nodes,
  strains, stresses and reactions. This function is substituted by the system of outdriver.
  ???!!! canidate for removal

  @param out - pointer to the opened text file
  @param lcid - load case id

  @return The function does not return anything.
  
  Created by JK,
*/
void lhsrhs::output (FILE *out,long lcid)
{
  print_displacements (out,lcid);
  if (Mp->straincomp==1)  print_strains (out,lcid);
  if (Mp->stresscomp==1)  print_stresses (out,lcid);
  if (Mp->reactcomp==1)   print_reactions (out,lcid);
}



/**
  Function reads initial conditions.
   
  @param in - pointer to the opened XFILE

  @return The function does not return anything.

  Created by JK, 17.1.2002
*/
void lhsrhs::initcond (XFILE *in)
{
  long i,j,k,l,ndofn;
  double v;

  if (Mp->tprob == forced_dynamics && Mb->dlc[0].tdl!=responsespectrum){
    for (i=0;i<Mt->nn;i++){
      ndofn=Mt->give_ndofn (i);
      xfscanf (in,"%ld",&l);
      
      //  nodal values
      for (j=0;j<ndofn;j++){
	k=Gtm->give_dof (l-1,j) - 1;
	xfscanf (in,"%lf",&v);
	if (k>-1)  lhs[k]=v;
      }
      
      //  time derivatives of nodal values
      for (j=0;j<ndofn;j++){
	k=Gtm->give_dof (l-1,j) - 1;
	xfscanf (in,"%lf",&v);
	if (k>-1)  tdlhs[k]=v;
      }
      
    }
  }
  if ((Mp->tprob == mat_nonlinear_statics) && Mb->nico)
  {
    for (i=0;i<Mt->nn;i++) 
    {
      ndofn=Mt->give_ndofn (i);
      //  nodal values
      for (j=0;j<ndofn;j++)
      {
	k=Gtm->give_dof (i,j) - 1;
	if ((k>-1) && (Mb->ico[i].type & inidisp))  
          lhs[k]=Mb->ico[i].val[j];
      }            
    }
  }

  /*
  if (Mp->tprob == forced_dynamics && Mb->dlc[0].tdl!=responsespectrum){
    for (i=0;i<Mt->nn;i++){
      ndofn=Mt->give_ndofn (i);
      
      //  nodal values
      for (j=0;j<ndofn;j++){
	k=Gtm->give_dof (i,j) - 1;
	if (k>-1)  lhs[k]=0.0;
      }
      
      //  time derivatives of nodal values
      for (j=0;j<ndofn;j++){
	k=Gtm->give_dof (i,j) - 1;
	if (k>-1)  tdlhs[k]=0.0;
      }
      
    }
  }
  */
}



/**
  Function cleans %vector of unknowns.
   
  @return The function does not return anything.

  Created by JK,
*/
void lhsrhs::clean_lhs ()
{
  long i;
  for (i=0;i<ndof;i++){
    lhs[i]=0.0;
  }
}

