#include <math.h>
#include "loadcase.h"
#include "global.h"
#include "globmat.h"
#include "elemswitch.h"
#include "gtopology.h"
#include "matrix.h"
#include "vector.h"
#include "loadn.h"
#include "loadel.h"
#include "element.h"
#include "node.h"
#include "elemhead.h"



/**
  Constructor initializes data members to zero or default values.

  Created by JK,
*/
loadcase::loadcase (void)
{
  //  number of loaded nodes
  nln = 0;
  //  number of loaded elements
  nle = 0;
  //  number of prescribed displacements
  npd = 0;
  //  number of prescribed temperature changes
  npt = 0;
  tempchang=0;
  
  //  loaded nodes
  lon = NULL;
  //  loaded elements
  loe = NULL;
  //  prescribed displacements
  pd = NULL;
  //  prescribed temperature changes
  pt = NULL;
}



/**
  Destructor releases allocated memory of the loadcase object.

  Created by JK,
*/
loadcase::~loadcase (void)
{
  delete [] lon;  delete [] loe;  delete [] pd;  delete [] pt;
}



/**
  Function reads characteristics of static load case from the opened 
  text file.

  @param in - pointer to the opened XFILE
   
  @return The function does not return anything.

  Created by JK, 24.7.2001
*/
void loadcase::read (XFILE *in)
{
  long i;
  
  //  loaded nodes
  xfscanf (in,"%ld",&nln);
  fprintf (stdout,"\n number of loaded nodes %ld",nln);
  lon = new loadn [nln];
  for (i=0;i<nln;i++){
    lon[i].read (in);
  }
  
  //  loaded elements
  xfscanf (in,"%ld",&nle);
  fprintf (stdout,"\n number of loaded elements %ld",nle);
  loe = new loadel [nle];
  for (i=0;i<nle;i++){
    loe[i].read (in);
  }
  
  //  prescribed displacements
  xfscanf (in,"%ld",&npd);
  fprintf (stdout,"\n number of prescribed displacements %ld",npd);
  pd = new double [npd];
  for (i=0;i<npd;i++){
    xfscanf (in,"%lf",pd+i);
  }
  
  //  prescribed temperature changes
  xfscanf (in,"%ld",&tempchang);
  if (tempchang==0)
    fprintf (stdout,"\n no temperature changes are described");
  if (tempchang>0)
    Mp->temperature=1;
  
  if (tempchang==1 || tempchang==2){
    npt = Mt->nn;
    pt = new double [npt];
    for (i=0;i<npt;i++){
      xfscanf (in,"%lf",pt+i);
    }
    
  }
  
  //if (tempchang==3)
  //rhstemp = new double [Ndofm];
  
}



/**
  Function assembles %vector of right hand side. 
  It is created by prescribed displacements, forces, moments, eigenstrains, etc.   
  In case that the flv is not NULL, %vector of load caused by forces only is stored in flv.

  @param lcid  - load case id
  @param rhs   - pointer to the right hand side
  @param flv   - array of load %vector caused by forces only (default value is NULL)
  @param scale - scale factor, it is used for subload cases
   
  Created by JK, 24.7.2001
  Modifed by Tomas Koudelka, 7.2008
*/
void loadcase::assemble (long lcid,double *rhs,double *flv,double scale)
{
  long i,j,k,l,ndofn,ndofe,*cn;
  vector r,f;
  matrix sm;
  
  //  array rhs and flv have to be cleaned in the caller (above) function
  
  //  contributions from nodes
  for (i=0;i<nln;i++){
    ndofn=Mt->give_ndofn (lon[i].nid);
    for (j=0;j<ndofn;j++){
      k=Mt->give_dof (lon[i].nid,j);
      if (k<=0)  continue;
      if (k>0)  rhs[k-1]+=lon[i].f[j];
    }
  }
    
  //  contributions from elements
  for (i=0;i<nle;i++){
    l=loe[i].eid;
    if (Gtm->leso[l]==1){
      ndofe=Mt->give_ndofe (loe[i].eid);
      cn = new long [ndofe];
      Mt->give_code_numbers (loe[i].eid,cn);
      
      for (j=0;j<ndofe;j++){
	k=cn[j]-1;
	if (k<0)  continue;
	else  rhs[k]+=loe[i].nf[j];
      }
      delete [] cn;
    }
  }
  
  cmulv(scale,rhs,Ndofm);
  // assembling of load vector caused by forces only    
  if (flv)
    copyv (rhs,flv,Ndofm);

  //  contributions from prescribed displacements
  if ((npd>0) && (Mp->tprob != mat_nonlinear_statics)){
    long ne = Mt->ne;
    for (i=0;i<ne;i++){
      if (Mt->elements[i].prescdispl==1){
	ndofe=Mt->give_ndofe (i);
	allocm (ndofe,ndofe,sm);
	stiffmat (lcid,i,sm);
	
	allocv (ndofe,r);
	allocv (ndofe,f);
	cn = new long [ndofe];
	Mt->give_code_numbers (i,cn);
	elprdispl (lcid,i,r.a,cn,ndofe);
	mxv (sm,r,f);
	cmulv (-1.0,f);
	locglob (rhs,f.a,cn,ndofe);
	destrm (sm);
	delete [] cn;
	destrv (f);  destrv (r);
      }
    }
  }

  //  contributions from temperature changes
  tempercontrib (lcid,rhs,scale);
}



/**
  Function computes contributions from temperature changes
  to the right hand side (%vector of nodal forces).
   
  @param lcid - load case id
  @param rhs - pointer to the right hand side
  @param scale - scale factor, it is used for subload cases
   
  Created by JK, Tomas Koudelka, 21.11.2007
*/
void loadcase::tempercontrib (long lcid,double *rhs,double scale)
{
  //  contributions from temperature changes
  if (tempchang==1){
    //  contribution to the time-independent problems from temperatures
    //  read from input files
    
    //  type of eigenstrain is temperature strain
    Mm->est=tempstrain;

    //  approximation of nodal temperatures to integration points
    intpointval (pt,temperature,1.0);
    
    //  thermal material models compute temperature strains
    Mm->temprstrains (lcid);
    
    //  computation of nodal forces caused by temperature
    nodal_eigstrain_forces (lcid,rhs);
  }
  
  if (tempchang==2){
    //  contribution to the time-dependent problems from temperatures
    //  read from input file, temperature is changed by multiplication of
    //  read function
    
    //  type of eigenstrain is temperature strain
    Mm->est=tempstrain;

    //  approximation of nodal temperatures
    intpointval (pt,temperature,scale);
    
    //  thermal material models compute temperature strains
    Mm->temprstrains (lcid);
    
    nodal_eigstrain_forces (lcid,rhs);
  }

  if (tempchang==3){
    //  contribution to the time-dependent coupled problems from
    //  temperatures obtained from heat transfer (from TRFEL)
    
    //  type of eigenstrain is temperature strain
    Mm->est=tempstrain;

    //  thermal material models compute temperature strains
    Mm->temprstrains (lcid);
    
    //  computation of nodal forces caused by temperature
    nodal_eigstrain_forces (lcid,rhs);
  }
  
}



/**
  Function computes reactions.
   
  @param lcid - load case id

  @return The function does not return anything.

  Created by JK,
*/
void loadcase::compute_reactions (long lcid)
{
  long i,j,k,l,ii,nne,ndofn,nn,ne,ndofe,*cn;
  ivector nod;
  vector f,tf;
  
  nn=Mt->nn;
  for (i=0;i<nn;i++){
    if (Mt->nodes[i].react==1){
      ndofn=Mt->give_ndofn (i);
      for (k=0;k<ndofn;k++){
	Mt->nodes[i].r[k]=0.0;
      }
    }
  }
  
  ne=Mt->ne;
  for (i=0;i<ne;i++){
    if ((Mt->elements[i].react==1) && (Gtm->leso[i]==1)){
      ndofe=Mt->give_ndofe (i);
      allocv (ndofe,f);
      elem_internal_forces (i,lcid,f);
      
      /*      if (Mt->elements[i].presctemp==1){
        //  type of eigenstrain is temperature strain
        Mm->est=tempstrain;
	allocv (ndofe,tf);
	elem_eigstrain_forces (lcid,i,tf);
	subv (f,tf,f);
	destrv (tf);
	}*/
      nne=Mt->give_nne (i);
      allocv (nne,nod);
      Mt->give_elemnodes (i,nod);
      ii=0;
      for (j=0;j<nne;j++){
	ndofn=Mt->give_ndofn (nod(j));
	if (Mt->nodes[nod(j)].react==1){
	  for (k=0;k<ndofn;k++){
	    Mt->nodes[nod(j)].r[k]+=f[ii];
	    ii++;
	  }
	}
	else  ii+=ndofn;
      }
      destrv (f);
      destrv (nod);
    }
  }
  
  for (i=0;i<nle;i++){
    l=loe[i].eid;
    if ((Mt->elements[l].react==1) && (Gtm->leso[i]==1)){
      ndofe=Mt->give_ndofe (l);
      cn = new long [ndofe];
      Mt->give_code_numbers (l,cn);
      nne=Mt->give_nne (l);
      allocv (nne,nod);
      Mt->give_elemnodes (l,nod);
      ii=0;
      for (j=0;j<nne;j++){
	ndofn=Mt->give_ndofn (nod(j));
	if (Mt->nodes[nod(j)].react==1){
	  for (k=0;k<ndofn;k++){
	    Mt->nodes[nod(j)].r[k]-=loe[i].nf[ii];
	    ii++;
	  }
	}
	else  ii+=ndofn;
      }
      destrv (nod);
      delete [] cn;
    }
  }
  
  
  
  /*  if (Mp->eigstrains==1){
    
    for (i=0;i<ne;i++){
	if ((Mt->elements[i].react==1) && (Gtm->leso[i]==1)){
	te = Mt->give_elem_type (i);
	ndofe=Mt->give_ndofe (i);
	allocv (ndofe,f);
	
	nne=Mt->give_nne (i);
	allocv (nne,nod);
	Mt->give_elemnodes (i,nod);
        //  type of eigenstrain is eigenstrain
        Mm->est=eigstrain;
	elem_eigstrain_forces (lcid,i,f);
	ii=0;
	for (j=0;j<nne;j++){
	  ndofn=Mt->give_ndofn (nod(j));
	  if (Mt->nodes[nod(j)].react==1){
	    for (k=0;k<ndofn;k++){
	      Mt->nodes[nod(j)].r[k]-=f[ii];
	      ii++;
	    }
	  }
	  else  ii+=ndofn;
	}
	destrv(nod);
	destrv (f);
      }
    }
  }
  */
  
}

